import math
from typing import Literal
class Rectangle:
      """
      A class to represent a rectangle, allowing calculations of its perimeter, surface, 
      and visual representation.

      Attributes:
            length (int): The length of the rectangle.
            width (int): The width of the rectangle.

      Methods:
            __init__(length: int, width: int):
                  Initializes the rectangle with a given length and width.

            perimetre() -> int:
                  Returns the perimeter of the rectangle.

            surface() -> int:
                  Returns the surface area of the rectangle.

            get_length() -> int:
                  Retrieves the current length of the rectangle.

            get_width() -> int:
                  Retrieves the current width of the rectangle.

            set_length(new_length: int) -> None:
                  Sets a new length for the rectangle.

            set_width(new_width: int) -> None:
                  Sets a new width for the rectangle.
      """

      def __init__(self, length: int, width: int):
            """
            Initializes the rectangle with a given length and width.

            Args:
                  length (int): The length of the rectangle.
                  width (int): The width of the rectangle.
            """
            self.length = length
            self.width = width


      def perimetre(self) -> int:
            """
            Calculates and returns the perimeter of the rectangle.

            Returns:
                  int: The perimeter, calculated as 2 * (length + width).
            """
            return (self.length + self.width) * 2

      def surface(self) -> int:
            """
            Calculates and returns the surface area of the rectangle.

            Returns:
                  int: The surface area, calculated as length * width.
            """
            return self.length * self.width

      def get_length(self) -> int:
            """
            Retrieves the current length of the rectangle.

            Returns:
                  int: The current length.
            """
            return self.length

      def get_width(self) -> int:
            """
            Retrieves the current width of the rectangle.

            Returns:
                  int: The current width.
            """
            return self.width

      def set_length(self, new_length: int) -> None:
            """
            Sets a new length for the rectangle.

            Args:
                  new_length (int): The new length to set.
            """
            self.length = new_length

      def set_width(self, new_width: int) -> None:
            """
            Sets a new width for the rectangle.

            Args:
                  new_width (int): The new width to set.
            """
            self.width = new_width

class Paralelepiped(Rectangle):
      """
      A subclass of the Rectangle class to represent a parallelepiped, 
      which extends a rectangle into three dimensions by adding a height.

      Attributes:
            length (int): The length of the base rectangle.
            width (int): The width of the base rectangle.
            height (int): The height of the parallelepiped.

      Methods:
            __init__(length: int, width: int, height: int):
                  Initializes the parallelepiped with length, width, and height.

            Volume() -> int:
                  Calculates and returns the volume of the parallelepiped.
      """

      def __init__(self, length: int, width: int, height: int):
            """
            Initializes the parallelepiped with a given length, width, and height.
            Calls the Rectangle constructor to initialize length and width.

            Args:
                  length (int): The length of the base rectangle.
                  width (int): The width of the base rectangle.
                  height (int): The height of the parallelepiped.
            """
            Rectangle.__init__(self, length, width)
            self.height = height

      def Volume(self) -> int:
            """
            Calculates and returns the volume of the parallelepiped.

            Returns:
                  int: The volume, calculated as length * width * height.
            """
            return self.length * self.width * self.height

class Bank_account:
      """
      A class to represent a bank account, allowing basic operations such as deposits, 
      withdrawals, fee deductions, and displaying account details.

      Attributes:
            acc_number (str): The account number.
            name (str): The name of the account holder.
            balance (float): The balance of the account.

      Methods:
            Payout(input_money: float) -> None:
                  Deposits a specified amount into the account.

            Withdrawal(output_money: float) -> str:
                  Withdraws a specified amount from the account, if sufficient funds are available.

            Fees() -> None:
                  Deducts a 5% fee from the account balance.

            Display() -> None:
                  Displays the account details in a formatted output.
      """

      def __init__(self, acc_number: str, name: str, balance: float):
            """
            Initializes the bank account with an account number, account holder name, and balance.

            Args:
                  acc_number (str): The account number.
                  name (str): The name of the account holder.
                  balance (float): The initial balance of the account.
            """
            self.acc_number = acc_number
            self.name = name
            self.balance = balance

      def Payout(self, input_money: float) -> None:
            """
            Deposits a specified amount into the account.

            Args:
                  input_money (float): The amount to deposit.
            """
            self.balance += input_money

      def Withdrawal(self, output_money: float) -> str:
            """
            Withdraws a specified amount from the account if sufficient funds are available.

            Args:
                  output_money (float): The amount to withdraw.

            Returns:
                  str: A message indicating the success of the withdrawal or insufficient funds.
            """
            if output_money > self.balance:
                  return "You're too broke to withdraw this amount."
            else:
                  self.balance -= output_money
                  return f"You successfully withdrew {output_money}. Your balance now is: {self.balance:.2f}."

      def Fees(self) -> None:
            """
            Deducts a 5% fee from the account balance.
            """
            self.balance = self.balance * 95 / 100  # Equivalent to self.balance * (1 - 5%).

      def Display(self) -> None:
            """
            Displays the account details in a formatted output.
            """
            print(f"* {'–' * 39} *")
            print(f"* {'Account Name:':<18} {self.name.capitalize():>20} *")
            print(f"* {'Account Number:':<18} {self.acc_number:>20} *")
            print(f"* {'Balance:':<18} {self.balance:>20.2f} *")
            print(f"* {'–' * 39} *")

class Circle:
      """
      A class to represent a circle with properties and methods to calculate its surface area, 
      perimeter, and determine the position of a point relative to the circle.

      Attributes:
            a_coord (float): The x-coordinate of the circle's center.
            b_coord (float): The y-coordinate of the circle's center.
            radius (float): The radius of the circle.

      Class Attributes:
            π (float): The mathematical constant pi, used for calculations.

      Methods:
            Surface() -> float:
                  Calculates the surface area of the circle.

            Perimetre() -> float:
                  Calculates the perimeter of the circle.

            Point_position(x_coord: float, y_coord: float) -> None:
                  Determines whether a point lies inside, on, or outside the circle.
      """

      π = math.pi  # π = 3.141592653589793

      def __init__(self, a_coordinate: float = 0, b_coordinate: float = 0, radius: float = 1):
            """
            Initializes the circle with a center (a_coordinate, b_coordinate) and a radius.

            Args:
                  a_coordinate (float): The x-coordinate of the circle's center. Default is 0.
                  b_coordinate (float): The y-coordinate of the circle's center. Default is 0.
                  radius (float): The radius of the circle. Default is 1.
            """
            self.a_coord = a_coordinate
            self.b_coord = b_coordinate
            self.radius = radius

      def Surface(self) -> float:
            """
            Calculates the surface area of the circle.

            Returns:
                  float: The surface area, calculated as π * radius^2.
            """
            return Circle.π * self.radius ** 2

      def Perimetre(self) -> float:
            """
            Calculates the perimeter of the circle.

            Returns:
                  float: The perimeter, calculated as 2 * π * radius.
            """
            return Circle.π * self.radius * 2

      def Point_position(self, x_coord: float, y_coord: float) -> None:
            """
            Determines whether a point lies inside, on, or outside the circle.

            Args:
                  x_coord (float): The x-coordinate of the point.
                  y_coord (float): The y-coordinate of the point.

            Prints:
                  str: A message indicating the position of the point relative to the circle.
            """
            distance = math.sqrt((x_coord - self.a_coord) ** 2 + (y_coord - self.b_coord) ** 2)
            if distance == self.radius:
                  print(f"The point M{(x_coord, y_coord)} lies on the circle Ϲ(O, {self.radius}).")
            elif distance < self.radius:
                  print(f"The point M{(x_coord, y_coord)} lies inside the circle Ϲ(O, {self.radius}).")
            else:
                  print(f"The point M{(x_coord, y_coord)} lies outside the circle Ϲ(O, {self.radius}).")
                  
class Calculation(int):
      """
      A class to perform various mathematical calculations including factorials, 
      summations, primality tests, and more.

      Methods:
            Factorial(integer: int) -> int:
                  Calculates the factorial of self.

            Sum(integer: int) -> int:
                  Calculates the sum of all integers from 1 to the given integer.

            testPrim(integer: int) -> bool:
                  Tests whether self is a prime number.

            testPrims(self: int, integer: int) -> str:
                  Checks if two integers are coprime (relatively prime).

            tableMult(integer: int) -> None:
                  Prints the multiplication table for self.

            allTablesMult() -> None:
                  Prints the multiplication tables for all integers from 1 to 9.

            listDiv(integer: int) -> list:
                  Returns a list of all divisors of self (excluding 1 and the integer itself).

            listDivPrim(integer: int) -> list:
                  Returns a list of all prime divisors of self.
      """

      def factorial(self) -> int:
            """
            Calculates the factorial of self.

            Returns:
                  int: The factorial of self.
            """
            facto = 1
            for i in range(2, self + 1):
                  facto *= i
            return facto

      def Sum(self) -> int:
            """
            Calculates the sum of all integers from 0 to self.

            Returns:
                  int: The sum of integers from 0 to the given number.
            """
            summ = 0
            for i in range(1, self + 1):
                  summ += i
            return summ

      def testPrim(self) -> bool:
            """
            Tests whether self is a prime number.

            Returns:
                  bool: True if the number is prime, False otherwise.
            """
            x = 0
            for i in range(2, self):
                  if self % i == 0:
                        x += 1
            return x == 0

      def testPrims(self, integer: int) -> str:
            """
            Checks if two integers are coprime (relatively prime).

            Args:
                  integer (int): The second number.

            Returns:
                  bool: True if they are coprime, False otherwise
            """
            PGCD = math.gcd(self, integer)
            if PGCD == 1:
                  return True
            else:
                  return False

      def listDiv(self) -> list:
            """
            Returns a list of all divisors of self (excluding 1 and the integer itself).

            Returns:
                  list: A list of divisors.
            """
            Ldiv = [1]
            for i in range(2, self):
                  if self % i == 0:
                        Ldiv.append(i)
            Ldiv.append(self)
            return Ldiv

      def listDivPrim(self) -> list:
            """
            Returns a list of all prime divisors of self.

            Returns:
                  list: A list of prime divisors.
            """
            Ldivprim = []
            for i in range(2, self):
                  if self % i == 0 and self.testPrim(i):
                        Ldivprim.append(i)
            return Ldivprim

class Geometry:
      """
      A class to perform basic geometric calculations, such as distances, midpoints, 
      triangle perimeters, and checks for isosceles triangles.

      Methods:
            distance(M1: tuple, M2: tuple) -> float:
                  Calculates the Euclidean distance between two points M1 and M2.

            middle(M1: tuple, M2: tuple) -> tuple:
                  Calculates the midpoint between two points M1 and M2.

            trianglePerimeter(M1: tuple, M2: tuple, M3: tuple) -> float:
                  Calculates the perimeter of a triangle formed by three points.

            triangleIsoscel(M1: tuple, M2: tuple, M3: tuple) -> bool:
                  Checks whether a triangle formed by three points is isosceles.
      """

      def __init__(self):
            """
            Initializes the Geometry class with no attributes, as it contains only utility methods.
            """
            pass

      def distance(self, M1: tuple, M2: tuple) -> float:
            """
            Calculates the Euclidean distance between two points M1 and M2.

            Args:
                  M1 (tuple): Coordinates of the first point, as (x1, y1).
                  M2 (tuple): Coordinates of the second point, as (x2, y2).

            Returns:
                  float: The distance between M1 and M2.
            """
            return math.sqrt((M1[0] - M2[0])**2 + (M1[1] - M2[1])**2)

      def middle(self, M1: tuple, M2: tuple) -> tuple:
            """
            Calculates the midpoint between two points M1 and M2.

            Args:
                  M1 (tuple): Coordinates of the first point, as (x1, y1).
                  M2 (tuple): Coordinates of the second point, as (x2, y2).

            Returns:
                  tuple: The midpoint between M1 and M2, as (x_mid, y_mid).
            """
            return ((M1[0] + M2[0]) / 2, (M1[1] + M2[1]) / 2)

      def trianglePerimeter(self, M1: tuple, M2: tuple, M3: tuple) -> float:
            """
            Calculates the perimeter of a triangle formed by three points.

            Args:
                  M1 (tuple): Coordinates of the first vertex, as (x1, y1).
                  M2 (tuple): Coordinates of the second vertex, as (x2, y2).
                  M3 (tuple): Coordinates of the third vertex, as (x3, y3).

            Returns:
                  float: The perimeter of the triangle.
            """
            M1_M2 = self.distance(M1, M2)
            M1_M3 = self.distance(M1, M3)
            M2_M3 = self.distance(M2, M3)
            return M1_M2 + M1_M3 + M2_M3

      def triangleIsoscel(self, M1: tuple, M2: tuple, M3: tuple) -> bool:
            """
            Checks whether a triangle formed by three points is isosceles.

            Args:
                  M1 (tuple): Coordinates of the first vertex, as (x1, y1).
                  M2 (tuple): Coordinates of the second vertex, as (x2, y2).
                  M3 (tuple): Coordinates of the third vertex, as (x3, y3).

            Returns:
                  bool: True if the triangle is isosceles, False otherwise.
            """
            M1M2 = self.distance(M1, M2)
            M1M3 = self.distance(M1, M3)
            M2M3 = self.distance(M2, M3)
            return M1M2 == M1M3 or M1M2 == M2M3 or M1M3 == M2M3

class Article:
      """
      Article Class:
      
      Represents a commercial article with a unique reference, description, and pricing details.

      Attributes:
            Reference (str): A unique identifier for the article (e.g., a code or ID).
            Designation (str): The name or description of the article.
            PrixHT (float): The price of the article excluding TVA (value-added tax).
            tauxTVA (float): The TVA rate applied to the article (default is 20%).
      """
      tauxTVA = 20 / 100  # Default TVA rate: 20%

      def __init__(self, Reference: str = "", Designation: str = "", PrixHT: float = 0.0):
            """
            Initializes an Article instance.

            Args:
                  Reference (str): The unique reference of the article. Default is an empty string.
                  Designation (str): The name or title of the article. Default is an empty string.
                  PrixHT (float): The price excluding TVA. Default is 0.0.
            """
            self.__reference = Reference
            self.__designation = Designation
            self.__prixHT = PrixHT

      @property
      def reference(self):
            """
            Gets the article's reference.
            """
            return self.__reference

      @reference.setter
      def reference(self, ref):
            """
            Sets the article's reference.
            """
            self.__reference = ref

      @property
      def designation(self):
            """
            Gets the article's designation.
            """
            return self.__designation

      @designation.setter
      def designation(self, desi):
            """
            Sets the article's designation.
            """
            self.__designation = desi

      @property
      def prixHT(self):
            """
            Gets the article's price excluding TVA.
            """
            return self.__prixHT

      @prixHT.setter
      def prixHT(self, prHT):
            """
            Sets the article's price excluding TVA.
            """
            self.__prixHT = prHT

      @property
      def tauxTVA(self):
            """
            Gets the current TVA rate.
            """
            return Article.tauxTVA

      @tauxTVA.setter
      def tauxTVA(self, TVA):
            """
            Sets a new TVA rate.
            """
            Article.tauxTVA = TVA

      def CalculerPrixTTC(self) -> float:
            """
            Calculates the total price including TVA.

            Returns:
                  float: The price including TVA.
            """
            return self.__prixHT * (1 + self.tauxTVA)

      def ShowArticle(self):
            """
            Displays the article's details, including reference, designation, and pricing.
            """
            print(f"Reference:     {self.__reference}")
            print(f"Designation:   {self.__designation}")
            print(f"PrixHT:        {self.__prixHT:.2f} DH")
            print(f"PrixTTC:       {self.CalculerPrixTTC():.2f} DH")

class Advance_str(str):
      """
      Advanced String Class:
      
      Extends the standard string class with additional functionality.

      Methods:
            __sub__(self1: str) -> Advance_str:
                  Removes the first occurrence of a given substring from the string.

            __mul__(listo: list[str], position: Literal["before", "after"] = "before") -> list[str]:
                  Combines the string with elements of a list, either before or after each element.

            append(str_: str) -> Advance_str:
                  Appends a given string to the current string.
      """

      def __str__(self):
            """
            Overrides the default string representation.
            """
            return super().__str__()

      def __sub__(self, self1):
            """
            Removes the first occurrence of the specified substring.

            Args:
                  self1 (str): The substring to be removed.

            Returns:
                  Advance_str: The modified string with the substring removed.
            """
            return Advance_str(self.replace(self1, "", 1))

      def __mul__(self, listo: list[str], position: Literal["before", "after"] = "before") -> list[str]:
            """
            Combines the string with elements from a list, either before or after each element.

            Args:
                  listo (list[str]): A list of strings.
                  position (Literal["before", "after"]): Whether to place the string before or after each list element.
                  Default is "before".

            Returns:
                  list[str]: A list of combined strings.
            """
            if position == "before":
                  return [self + " " + f for f in listo]
            elif position == "after":
                  return [f + " " + self for f in listo]

      def append(self, str_: str) -> 'Advance_str': 
            """
            Appends a given string to the current string.

            Args:
                  str_ (str): The string to append.

            Returns:
                  Advance_str: The modified string with the appended value.
            """
            return Advance_str(self + " " + str_)
